'''
    Project: Gin Rummy
    File name: gin_rummy/utils/gin_rummy_error.py
    Author: William Hale
    Date created: 4/29/2020
'''


class GinRummyError(Exception):
    pass


class GinRummyProgramError(GinRummyError):
    pass
